/***************************************************************************************************************************
Title: Lego Head Safe
Author: Kaitlin Bellerose
Date: 11/01/2014
Updated: 12/03/2014

Purpose: This sketch provides the control system for the Lego Head safe.

This sketch borrows from multiple sources.  The original sources will be cited here in the title block when general
approaches were followed.

1) Keypad.h library written by Mark Stanley and Alexander Brevig (mstanley@technologist.com, alexanderbrevig@gmail.com)
2) There is a password.h library available on Instructables and similar sites.  I did not care for the implementation, so
   I created my own.
****************************************************************************************************************************/

#include <Wire.h> 
#include <LiquidCrystal_I2C.h>
#include <Servo.h>
#include <Keypad.h>

LiquidCrystal_I2C lcd(0x27,16,2);       // Instantiate a 16x2 I2C LCD display with address 0x27

Servo myServo;                          // Instantiate a servo
int pos;                                // Declare variable to hold servo position value (NOTE: "position is a reserved word)
int servoSpeed;                         // Declare variable to hold servo speed value (NOTE: "speed" is a reserved word)
char password[3];                       // Array that holds master password.  Set to length you would like.  I chose 3 for ease of demo
char entry[3];                          // Array that holds the user-entered password attempt
boolean isAlreadyOpen;                  // Flag that tracks whether safe is already open 
boolean isAlreadyClosed;                // Flag that tracks whether safe is already closed


const byte ROWS = 4;                    // Four rows in 4x4 keypad matrix
const byte COLS = 4;                    // Four columns in 4x4 keypad matrix
char keys[ROWS][COLS] = {               // Maps the keypad values
  {'1','2','3', 'A'},
  {'4','5','6', 'B'},
  {'7','8','9', 'C'},
  {'*','0','#', 'D'}
};
byte rowPins[ROWS] = {8, 7, 6, 5};    // Connect to the row pinouts of the keypad
byte colPins[COLS] = {4, 3, 2, 10};   // Connect to the column pinouts of the keypad
                                      // Use 1 or more appropriate pin for what I have labeled as "10"
                                      // My Arduino has many dead ports I have to navigate around

Keypad keypad = Keypad( makeKeymap(keys), rowPins, colPins, ROWS, COLS );  // Instantiates a keypad with the values passed.


/*******************************Set-up Function********************************************/

void setup(){
  Serial.begin(9600);                    // Set serial to 9600 baud
  myServo.attach(11);                    // Set servo to pin 11
  
  isAlreadyOpen = 0;                     // Sets open flag to false
  isAlreadyClosed = 1;                   // Sets closed flag to true
  
  pos = 90;                              // Sets value for servo arm position at 90 degrees
  servoSpeed = 3;                        // Sets servo speed to 3 degrees per second
  myServo.write(pos);                    // Sets servo arm to 90 degrees
  
  password[0] = '1';                     // Be sure to put initial value in single quotes due to its char nature
  password[1] = '2';                     // Second value in password array
  password[2] = '3';                     // Last value in password.  The password will default to "123" when
                                         // power is lost.
  
  keypad.setDebounceTime(250);           // Sets the button debounce time for the keypad
  
  lcd.init();                            // Initialize the LCD 
  lcd.backlight();                       // Set the display to be backlit
  lcd.print("Welcome Home!");            // Print welcome message to LCD
  delay(1000);                           // Have message display for 1000 milseconds
  lcd.clear();                           // Clear the display
  lcd.print("1) Open 2) Close");         // Print menu options
  lcd.setCursor(0,1);                    // Move cursor to second line
  lcd.print("3) Set Code");              // Print second part of menu options
  delay(3000);                           // Delay
}

/***********************************Main Loop**********************************************/

void loop(){

  
  entry[0] = NO_KEY;                      // Reset the entry array with each loop or else
  entry[1] = NO_KEY;                      // you'll get "false" entries from existing values
  entry[2] = NO_KEY;                      // Set the final keypad to NO_KEY (similar to setting 
                                          // value to NULL in C++)
  
  char key = keypad.getKey();             // Get the first key press with regard to menu.
  
  if (key == '1'){                        // Initiates logic if "Open safe" option is selected
     if(isAlreadyOpen == 1)               // Checks to see if the safe is already open.  No point in opening something already open
      {
        lcd.clear();                      // Clear LCD
        lcd.print("Safe is already");     // Prints message 
        lcd.setCursor(0,1);               // Move cursor to second line
        lcd.print("open.");               // Print second part of menu options
        delay(2000);                      // Delay 2000 miliseconds
      }
      else                                // If the safe is not already open, the following code engages
      {
        lcd.clear();                      // Clear the LCD
        lcd.print ("Enter 1st key:");     // Message to solicit the first key press
        lcd.setCursor(0,1);               // Moves cursor to second line
         
        while (entry[0] == NO_KEY)        // The loop runs until a valid key press is made
        {
          entry[0] = keypad.getKey();     // Stores the first value of the user-entered attempt at password
        }
        lcd.print(entry[0]);              // Prints value for display on LCD.  Use for debugging and ease of use.
        delay(1000);                      // Delays for 1000 miliseconds
        lcd.clear();                      // Clears LCD
    
        lcd.print("Enter 2nd key:");      // Enter the second key press
        lcd.setCursor(0,1);               // Move curosr
    
        while (entry[1] == NO_KEY)        // Another loop that will run endlessly unless a valid keypress is made
        {
          entry[1] = keypad.getKey();     // Stores the second value of the user-entered attempt at password
        }
        lcd.print(entry[1]);              // Prints value for display on LCD.  Use for debugging and ease of use.
        delay(1000);                      // Delay for 1000 miliseconds
        lcd.clear();                      // Clear LCD
    
        lcd.print("Enter 3rd key:");      // Solicit third key press
        lcd.setCursor(0,1);               // Moved the cursor
        
        while (entry[2] == NO_KEY)        // Again, a while loop to continuously poll unti a valid key press is made.
        {
          entry[2] = keypad.getKey();     // Gets the final key press
        }
        lcd.print(entry[2]);              // Prints value for display on LCD.  Use for debugging and ease of use.
        delay(1000);                      // Delay for 1000 miliseconds
        lcd.clear();                      // Clear LCD
    
        if ((entry[0] == password[0]) && entry[1] == password[1] && entry[2] == password[2])  // If what the user entered matches the stored password, this loop executes
        {
     
            isAlreadyOpen = 1;            // Set isAlreadyOpen flag to true/high
            isAlreadyClosed = 0;          // Set isAlreadyClosed flag to false/low
            pos = 180;                    // Set position to 180 degrees
            myServo.write(pos);           // Set the servo to 180 degrees
            lcd.clear();                  // Clear LCD in anticipation of message
            lcd.print("Safe is open.");   // Message that the safe is open
            delay(2000);                  // Delay of 2000 miliseconds
        }
       else                               // If the wrong code was entered, this executes
       {
          lcd.clear();                    // Clear LCD
          lcd.print("Invalid code.");     // Invalid code message
          delay(1000);                    // Delay 1000 miliseconds
          lcd.clear();                    // Clear LCD
       }
  }
    mainLCDMenu();                        // Regardless of whether password entry was successful, this kicks back to the main menu functions
      
 }
  else if (key == '2'){                   // Initiates if "Close safe" is selected
    if(isAlreadyClosed == 1)              // If the heaad is already closed, why close it again?
    {
      lcd.clear();                        // Clear LCD
      lcd.print("Safe is already");       // Begin message that the safe is already closed
      lcd.setCursor(0,1);                 // Move cursor to second line
      lcd.print("closed.");               // Finish message
      delay(2000);                        // Delay for 2000 miliseconds delay
    }
    else                                  // If the safe is not already closed, this executes
    {
      isAlreadyOpen = 0;                  // Sets isAlreadyOpen flag to false/low
      isAlreadyClosed = 1;                // Sets isAlreadyClosed flag to true/high
      pos = 90;                           // Sets pos to 90 degrees
      myServo.write(pos);                 // Sets servo to 90 degrees
      lcd.clear();                        // Clears LCD
      lcd.print("Safe is closed.");       // Message that the safe is closed
      delay(2000);                        // Delay 2000 miliseconds
     }
     lcd.clear();                         // Clear LCD
     mainLCDMenu();                       // Either way, go back to main menu
  }
else if (key == '3'){                     // If "Set code" option is selected 
  lcd.clear();                            // Clear LCD
  
    lcd.print ("Enter 1st key:");         // The code that follow ia a repeat of code in option 1
    lcd.setCursor(0,1);                   // The user has to enter the password successfully before changing it
    while (entry[0] == NO_KEY)
    {
      entry[0] = keypad.getKey();
    }
    lcd.print(entry[0]);
    delay(1000);
    lcd.clear();
    
    lcd.print("Enter 2nd key:");
    lcd.setCursor(0,1);
    
    while (entry[1] == NO_KEY)
    {
      entry[1] = keypad.getKey();
    }
    lcd.print(entry[1]);
    delay(1000);
    lcd.clear();
    
    lcd.print("Enter 3rd key:");
    lcd.setCursor(0,1);
    
    while (entry[2] == NO_KEY)
    {
      entry[2] = keypad.getKey();
    }
    lcd.print(entry[2]);
    delay(1000);
    lcd.clear();
    
    if ((entry[0] == password[0]) && entry[1] == password[1] && entry[2] == password[2])  // If the password was correctly entered, this logic executes
    {
        password[0] = NO_KEY;                  // Reset the password value 0 to NO-KEY
        password[1] = NO_KEY;                  // Reset the password value 1 to NO-KEY
        password[2] = NO_KEY;                  // Reset the password value 2 to NO-KEY
        lcd.print ("Set 1st key:");            // Set first key value
        lcd.setCursor(0,1);
 
        while (password[0] == NO_KEY)          // Loops endlessly until valid key press received
        {
           password[0] = keypad.getKey();      // Sets password[0] to new value
        }  
        lcd.print(password[0]);                // Prints out new value
        delay(1000);
        lcd.clear();                           // Clear LCD
    
         lcd.print("Set 2nd key:");            // Message for 2nd key
         lcd.setCursor(0,1);                   // Move code
    
         while (password[1] == NO_KEY)         // Loops endlessly until valid key press received
        {
          password[1] = keypad.getKey();       // Sets password[1] to new value
        }
        lcd.print(password[1]);                // Prints out new value
        delay(1000);                           // Delay 1000 miliseconds
        lcd.clear();                           // Clear LCD
    
        lcd.print("Set 3rd key:");             // Message for third key
        lcd.setCursor(0,1);                    // Move cursor
         
        while (password[2] == NO_KEY)          // Loops endlessly until valid key press received
        {
          password[2] = keypad.getKey();       // Sets password[2] to new value
        }
        lcd.print(password[2]);                // Prints out new value
        delay(1000);                           // Delay 1000 miliseconds
        lcd.clear();                           // Clear LCD
    }
    
    else                                      // This executes if the user enters the wrong password for verification prior
    {                                         // to changing the password.  User cannot change password without knowing the password
      lcd.clear();                            // Clear LCD
      lcd.print("Invalid code.");             // Invalid code message
      delay(1000);                            // Delay 1000 miliseconds
      lcd.clear();                            // Clear LCD
    }
    mainLCDMenu();                            // Regardless of outcome, go back to the main menu
  }
}
